/*
 * localCPU.cpp
 *
 *  Created on: 2010-11-06
 *      Author: Alex Krizhevsky (akrizhevsky@gmail.com)
 */
#include <math.h>
#include "../include/localCPU.h"


/*
 * hidActs:     (numModules, numFilters, numImages) if hidActsOrder == MODULE_FILTER_IMAGE
 *              (numFilters, numModules, numImages) otherwise
 * filters:     (numModules, numColors, filterPixels, numFilters)
 *
 * targets:     (numColors, imgPixels, numImages)
 */
void cpuImgActs(float* hidActs, float* filters, float* targets,
               int numModulesX,  int numImages,  int numFilters,
               int filterSize,  int imgSize,  int moduleStart,  int moduleStride,  int numColors, FILTER_OUTPUT_ORDER hidActsOrder) {
    int filterPixles = filterSize * filterSize;
    int imgPixels = imgSize * imgSize;
    int numModules = numModulesX * numModulesX;
    for (int py = 0; py < imgSize; py++) {
        for (int px = 0; px < imgSize; px++) {
            for (int my = 0; my < numModulesX; my++) {
                int moduleTop = moduleStart + my * moduleStride;
                int moduleBottom = moduleTop + filterSize;
                for (int mx = 0; mx < numModulesX; mx++) {
                    int m = my * numModulesX + mx;
                    int moduleLeft = moduleStart + mx * moduleStride;
                    int moduleRight = moduleLeft + filterSize;
                    int pixInModuleX = px - moduleLeft;
                    int pixInModuleY = py - moduleTop;
                    int pixInModule = pixInModuleY * filterSize + pixInModuleX;
                    if (py >= moduleTop && py < moduleBottom && px >= moduleLeft && px < moduleRight) {
                        for (int f = 0; f < numFilters; f++) {
                            for (int i = 0; i < numImages; i++) {
                                for (int c = 0; c < numColors; c++) {
                                    float w = filters[(m*numColors+c) * numFilters * filterPixles + pixInModule * numFilters + f];
                                    float h = hidActsOrder == MODULE_FILTER_IMAGE 
                                            ? hidActs[m * numImages * numFilters + f * numImages + i]
                                            : hidActs[m * numImages + f * numModules * numImages + i];
                                    targets[c * imgPixels * numImages + i] += w * h;
                                }
                            }

                        }
                    }
                }
            }
            targets += numImages;
        }
    }
}

/*
 * images:      (numColors, imgPixels, numImages) with stride given
 * filters:     (numModules, numColors, filterPixels, numFilters)
 *
 * targets:     (numModules, numFilters, numImages) if foo == MODULE_FILTER_IMAGE
 *              (numFilters, numModules, numImages) otherwise
 */
void cpuFilterActs(float* images, float* filters, float* targets,
                       int numImages, int numFilters,
                       int imgSize, int filterSize, int paddingStart,
                       int moduleStride, int numModulesX, int numColors, FILTER_OUTPUT_ORDER targetsOrder) {
    int filterPixels = filterSize * filterSize;
    int numModules = numModulesX * numModulesX;
    int imgPixels = imgSize * imgSize;
    for (int my = 0; my < numModulesX; my++) {
        int mStartY = paddingStart + my * moduleStride;
        for (int mx = 0; mx < numModulesX; mx++) {
            int mStartX = paddingStart + mx * moduleStride;
            int m = (my * numModulesX + mx);
            for (int f = 0; f < numFilters; f++) {
                for (int i = 0; i < numImages; i++) {
                    float prod = 0;
                    for (int c = 0; c < numColors; c++) {
                        for (int y = 0; y < filterSize; y++) {
                            for (int x = 0; x < filterSize; x++) {
                                float imgVal = mStartY + y >= 0 && mStartY + y < imgSize && mStartX + x >= 0 && mStartX + x < imgSize
                                            ? images[c * imgPixels * numImages + i + ((mStartY+y) * imgSize + mStartX+x) * numImages]
                                            : 0;
                                prod += filters[c * filterPixels * numFilters + m * numFilters * filterPixels * numColors + f + (y * filterSize + x) * numFilters] * imgVal;
                            }
                        }
                    }
                    if (targetsOrder == MODULE_FILTER_IMAGE) {
                        targets[m * numFilters * numImages + f * numImages + i] = prod;
                    } else {
                        targets[f * numModules * numImages + m * numImages + i] = prod;
                    }
//                    targets++;
                }
            }
        }
    }
}

/*
 * images:      (numColors, imgPixels, numImages), with stride given
 * hidActs:     (numModules, numFilters, numImages) if foo == MODULE_FILTER_IMAGE
 *              (numFilters, numModules, numImages) otherwise
 *
 * targets:     (numModules, numColors, filterPixels, numFilters)
 */
void cpuWeightActs(float* images, float* hidActs, float* targets,
                   int numImages,  int numFilters, int imgSize,  int filterSize,
                   int numModulesX, int paddingStart, int moduleStride, int numColors, FILTER_OUTPUT_ORDER hidActsOrder) {
    int filterPixels = filterSize * filterSize;
    int imgPixels = imgSize * imgSize;
    int numModules = numModulesX * numModulesX;
    for (int i = 0; i < numImages; i++) {
        for (int my = 0; my < numModulesX; my++) {
            int mStartY = paddingStart + my * moduleStride;
            for (int mx = 0; mx < numModulesX; mx++) {
                int mStartX = paddingStart + mx * moduleStride;
                int m = (my * numModulesX + mx);
                for (int f = 0; f < numFilters; f++) {
                    float h = hidActsOrder == MODULE_FILTER_IMAGE ? hidActs[m *numFilters*numImages + f * numImages + i] : hidActs[m *numImages + f * numImages*numModules + i];
                    for (int c = 0; c < numColors; c++) {
                        for (int y = 0; y < filterSize; y++) {
                            for (int x = 0; x < filterSize; x++) {
                                float imgVal = 0;
                                if (mStartY + y >= 0 && mStartY + y < imgSize && mStartX + x >= 0 && mStartX + x < imgSize) {
                                    imgVal = images[c * numImages * imgPixels + i + numImages * (imgSize * (mStartY + y) + mStartX + x)];
                                }
                                targets[c * filterPixels * numFilters + m * numFilters * filterPixels * numColors + f + (y * filterSize + x) * numFilters] += h * imgVal;
                            }
                        }
                    }
                }
            }
        }
    }
}
